
from copy import deepcopy

from itertools import islice
from collections import OrderedDict
from collections import namedtuple

from .. util.iter import OrderedDefaultDict

from .. util.iter import force_two_dimen
from .. util.iter import num_dimen
from .. util.iter import reduce_extra_dimen
from .. util.iter import transpose
from .. util.iter import index_sequence
from .. util.iter import make_iterable
from .. util.iter import ordered_unique
from .. util.iter import is_vengeance_class

from .. util.text import repr_iter
from .. util.text import p_json_dumps

from .. util.filesystem import write_file
from .. util.filesystem import read_file
from .. util.filesystem import apply_file_extension

from . flux_row_cls import flux_row_cls


class flux_cls:

    def __init__(self, matrix):
        self.headers = OrderedDict()

        self._num_cols = None
        self._matrix   = None

        self.__to_flux_rows(matrix)

    @property
    def header_values(self):
        return self._matrix[0].values

    @property
    def num_cols(self):
        """ find the maximum column length of all rows """
        if isinstance(self._num_cols, int):
            return self._num_cols

        self._num_cols = max(len(row) for row in self._matrix)

        return self._num_cols

    @property
    def is_jagged(self):
        num_cols = len(self._matrix[0])
        for row in self._matrix[1:]:
            if len(row) != num_cols:
                return True

        return False

    @property
    def is_empty(self):
        return self.num_rows == 0

    @property
    def num_rows(self):
        """ header row not included in count """
        return len(self._matrix) - 1

    def execute_commands(self, commands, profile_performance=False):
        if profile_performance:
            try:
                from line_profiler import LineProfiler
            except ImportError as e:
                raise ImportError('u need to install the line_profiler package') from e

            profiler = LineProfiler()
        else:
            profiler = None

        commands = self.__parse_commands(commands)

        for command in commands:
            method = command.method
            if profiler:
                method = profiler(method)

            if command.num_args == 0:
                method()
            elif command.num_args == 1:
                method(command.args)
            else:
                method(*command.args)

        if profiler:
            profiler.print_stats()

    def __parse_commands(self, commands):
        parsed = []
        command_names = index_sequence(['method_name', 'method', 'args', 'num_args'])

        for command in commands:
            if isinstance(command, (list, tuple)):
                name, args = command
                num_args = len(args)
            else:
                name = command
                args = None
                num_args = 0

            method = getattr(self, name)
            row_o  = flux_row_cls(command_names, [name, method, args, num_args])
            row_o.bind()

            parsed.append(row_o)

        return parsed

    def reapply_header_names(self, headers):
        """
        as to not de-reference flux_row_cls._headers in meta_m

        can NOT reset self.headers to new variable, each item must be individually reassigned
        """
        if len(headers) != self._num_cols:
            self._num_cols = None

        self.headers.clear()
        for i, header in enumerate(headers):
            self.headers[header] = i

    def matrix_by_headers(self, *headers):
        columns = transpose([row.values for row in self._matrix[1:]])

        m = []
        for header in reduce_extra_dimen(headers, 1):
            header = self.__process_header(header)

            if header not in self.headers:
                column = [None] * self.num_rows
            else:
                i = self.headers[header]
                column = columns[i]

            column.insert(0, header)
            m.append(column)

        m = transpose(m)
        self.__to_flux_rows(m)

    def __process_header(self, header):
        """
        inserted headers must be surrounded by parenthesis to ensure these
        new columns are being created intentionally (and not because of spelling errors, etc)
        """

        # renamed header
        if isinstance(header, dict):
            h_old, h_new = tuple(header.items())[0]
            self.headers[h_new] = self.headers.pop(h_old)

            return h_new

        # inserted header
        if header.startswith('(') and header.endswith(')'):
            return header.replace('(', '').replace(')', '')

        if header not in self.headers:
            raise ValueError("'{header}' not found in flux headers\n"
                             "inserted columns should be surrounded by parenthesis, ie '({header})' not '{header}'"
                             .format(header=header))

        return header

    def rename_columns(self, old_to_new_headers):
        """
        :param old_to_new_headers: a dictionary of {h_old: h_new} headers
        """
        renamed = self.header_values

        for h_old, h_new in old_to_new_headers.items():
            if h_old not in self.headers:
                raise KeyError('invalid header name {}'.format(h_old))

            i = self.headers[h_old]
            renamed[i] = h_new

        self.reapply_header_names(renamed)

    def insert_columns(self, *inserted):
        """
        :param inserted: (before_header, new_header)

        eg inserted:
            [(0,          'header_a'),      insert column at index 0
             ('*f',       'header_a'),      insert column at index 0
             ('header_a', 'header_b')]      insert column before 'header_b'
        """
        inserted = reduce_extra_dimen(inserted)
        nd = num_dimen(inserted)

        if nd == 0:
            inserted = [(0, inserted)]
        elif nd == 1:
            inserted = [inserted]

        headers = self.header_values
        for before, header in inserted:
            i = None

            if isinstance(before, int):
                i = before
            elif isinstance(before, str):
                if before == '*f':
                    i = 0
                elif before == '*l':
                    i = len(headers)
                else:
                    before = before.replace('*', '')
                    i = headers.index(before)

            if i is None:
                raise KeyError("insertion header '{}' does not exist".format(before))

            headers.insert(i, header)

        indeces = []
        for _, header in inserted:
            indeces.append(headers.index(header))

        indeces.sort()

        for i in indeces:
            for row in self:
                row.values.insert(i, None)

        self.reapply_header_names(headers)

    def append_columns(self, *header_names):
        header_names = reduce_extra_dimen(header_names, minimum=1)
        self._matrix[0].values.extend(header_names)

        for header in header_names:
            self.headers[header] = len(self.headers)
            for row in self:
                row.values.append(None)

    def delete_columns(self, *header_names):
        header_names = reduce_extra_dimen(header_names, minimum=1)
        indeces = [self.headers[h] for h in header_names]

        for c in sorted(indeces, reverse=True):
            for row in self._matrix:
                del row.values[c]

        self.reapply_header_names(self.header_values)

    def insert_rows(self, i, rows):
        if is_vengeance_class(rows):
            if i == 0:
                rows = list(rows.rows())
            else:
                rows = list(rows.rows())[1:]

        rows = force_two_dimen(rows)

        if i == 0:
            self.reapply_header_names(rows[0])
            del self._matrix[0]

        self._num_cols = None

        m = []
        for row in rows:
            if isinstance(row, flux_row_cls):
                row = row.values

            row = flux_row_cls(self.headers, row)
            m.append(row)

        self._matrix[i:i] = m

    def append_rows(self, rows):
        if is_vengeance_class(rows):
            rows = list(rows.rows())[1:]

        rows = force_two_dimen(rows)

        if len(self._matrix) == 0:
            self.headers = index_sequence(str(v) for v in rows[0])

        self._num_cols = None

        for row in rows:
            if isinstance(row, flux_row_cls):
                row = row.values

            row = flux_row_cls(self.headers, row)
            self._matrix.append(row)

    def fill_jagged_columns(self):
        num_cols = self.num_cols

        for row in self:
            num_missing_c = num_cols - len(row)
            if num_missing_c > 0:
                row.values.extend([None] * num_missing_c)

    def sort(self, *f, **kwargs):
        """ in-place
        eg:
            flux.sort('col_a', 'col_b', 'col_c',
                      reverse=[True, True, True])
        """
        self._matrix[1:] = self.__sort_rows(f, kwargs)

    def sorted(self, *f, **kwargs):
        """ returns new flux
        eg:
            flux = flux.sorted('col_a', 'col_b', 'col_c',
                               reverse=[True, True, True])
        """
        m = [row.values for row in self.__sort_rows(f, kwargs)]
        m.insert(0, self.header_values)

        return self.__class__(m.copy())

    def filter(self, f):
        """ in-place """
        self._matrix[1:] = [row for row in self if f(row)]

    def filtered(self, f):
        """ return new flux """
        m = [row.values for row in self if f(row)]
        m.insert(0, self.header_values)

        return self.__class__(m.copy())

    def filter_by_unique(self, *f):
        self.__filter_unique_rows(f, in_place=True)

    def filtered_by_unique(self, *f):
        return self.__filter_unique_rows(f, in_place=False)

    def __sort_rows(self, f, kwargs):
        if kwargs and 'reverse' not in kwargs:
            raise ValueError("only 'reverse' keyword is accepted")

        reverses = list(kwargs.get('reverse', []))

        for _ in range(len(f) - len(reverses)):
            reverses.append(False)

        m = [row for row in self]
        for f, reverse in zip(f, reverses):
            f = self.__row_values_accessor(f)
            m.sort(key=f, reverse=reverse)

        return m

    def __filter_unique_rows(self, f, in_place):
        uniq = set()

        def evaluate_unique(row):
            v = f(row)
            if v not in uniq:
                uniq.add(v)
                return True
            else:
                return False

        f = self.__row_values_accessor(f)
        if in_place:
            self.filter(evaluate_unique)
        else:
            return self.filtered(evaluate_unique)

    def unique_values(self, *f):
        """ :return: unique values within column(s) """
        f = self.__row_values_accessor(f)
        return ordered_unique(f(row) for row in self)

    def reset_matrix(self, m):
        self._matrix   = None
        self._num_cols = None

        self.__to_flux_rows(m)

    def index_row(self, *f):
        """ dictionary of {f(row): row}
        overwrites row values for non-unique key values
        """
        f = self.__row_values_accessor(f)
        return OrderedDict((f(row), row) for row in self)

    def index_rows(self, *f):
        """ dictionary of {f(row): [rows]}
        preserves rows in a list for non-unique key values
        """
        f = self.__row_values_accessor(f)
        items = [(f(row), row) for row in self]

        return OrderedDefaultDict(list, items)

    def namedtuples(self):
        nt_cls = namedtuple('flux_row_nt', self.header_values)
        return [nt_cls(*row.values) for row in self]

    def bind(self):
        """ speed up attribute access by binding values directly to row.__dict__ """
        [row.bind() for row in self._matrix]

    def to_json(self, path=None):
        j = [row.dict() for row in self]
        j = p_json_dumps(j)

        if path is None:
            return j

        write_file(path, j)

    @classmethod
    def from_json(cls, path):
        rows = read_file(path)
        if rows:
            m = [list(d.values()) for d in rows]
            m.insert(0, list(rows[0].keys()))
        else:
            m = [[]]

        return cls(m)

    def to_csv(self, path):
        m = list(self.rows())
        write_file(path, m)

    @classmethod
    def from_csv(cls, path):
        m = read_file(path)
        return cls(m)

    def serialize(self, path):
        path = apply_file_extension(path, '.flux')
        write_file(path, self)

    @classmethod
    def deserialize(cls, path):
        path = apply_file_extension(path, '.flux')
        return read_file(path)

    def rows(self, r_1='*h', r_2='*l'):
        r_1 = self.__matrix_index(r_1)
        r_2 = self.__matrix_index(r_2)

        return (row.values for row in islice(self._matrix, r_1, r_2 + 1))

    def flux_rows(self, r_1='*h', r_2='*l'):
        r_1 = self.__matrix_index(r_1)
        r_2 = self.__matrix_index(r_2)

        return iter(islice(self._matrix, r_1, r_2 + 1))

    def __to_flux_rows(self, m):
        if is_vengeance_class(m):
            m = m.rows()

        m = force_two_dimen(m)

        if 'values' in m[0]:
            raise NameError("'values' must not be in header row {}\n(causes conflict with flux_row_cls property)"
                            .format(m[0]))

        if 'names' in m[0]:
            raise NameError("'names' must not be in header row {}\n(causes conflict with flux_row_cls property)"
                            .format(m[0]))

        self.headers = index_sequence(str(v) for v in m[0])
        self._matrix = []

        for row in m:
            if not isinstance(row, list):
                raise ValueError('row must be list')

            row_o = flux_row_cls(self.headers, row)
            self._matrix.append(row_o)

    def __to_primitive_rows(self):
        if not isinstance(self._matrix[0], flux_row_cls):
            return

        self._matrix = [row.values for row in self._matrix]

    def __row_values_accessor(self, f):
        """ convert f into a function that can be called on each row in self._matrix """

        # return muliple values (as tuple) from each row
        def row_values(row):
            return tuple(row.values[c] for c in columns)

        # return single value from each row
        def row_value(row):
            return row.values[i]

        f = reduce_extra_dimen(f)
        self.__validate_row_accessor(f)

        if callable(f):
            return f

        if isinstance(f, (list, tuple)):
            columns = tuple(self.headers[n] for n in f)
            f = row_values
        elif isinstance(f, str):
            i = self.headers[f]
            f = row_value
        elif isinstance(f, int):
            i = f
            f = row_value

        return f

    def __validate_row_accessor(self, f):
        if isinstance(f, (str, list, tuple)):
            invalid = [h for h in make_iterable(f) if h not in self.headers]
            if invalid:
                raise KeyError("invalid column reference: {}".format(repr_iter(invalid)))

        elif isinstance(f, int):
            if f >= len(self.headers):
                raise KeyError("integer out of bounds: {}".format(f))

        else:
            raise TypeError("unhandled type conversion for '{}'".format(f))

    def __matrix_index(self, reference):
        r_i = None

        if isinstance(reference, (int, float)):
            r_i = int(reference)
        elif reference == '*h':
            r_i = 0
        elif reference == '*f':
            r_i = 1
        elif reference == '*l':
            r_i = self.num_rows

        if r_i is None:
            raise KeyError("invalid index '{}'".format(reference))

        return r_i

    def copy(self, deep=False):
        if deep:
            return deepcopy(self)

        # self._num_cols?
        m = [row.values for row in self._matrix]
        return self.__class__(m.copy())

    def __len__(self):
        return len(self._matrix)

    def __getitem__(self, ref):
        if isinstance(ref, int):
            return self._matrix[ref]

        if isinstance(ref, str):
            if ref not in self.headers:
                raise ValueError("column '{}' not in headers".format(ref))

            i = self.headers[ref]
            return [row.values[i] for row in self]

        if isinstance(ref, slice):
            i_1, i_2, step = ref.start, ref.stop, ref.step
            if i_1 is not None:
                if i_1 < 0:
                    i_1 = len(self._matrix) + i_1

            if i_2 is not None:
                if i_2 < 0:
                    i_2 = len(self._matrix) + i_2

            return list(row for row in islice(self._matrix, i_1, i_2, step))

        raise KeyError("reference syntax should be integers, eg 'flux[1:3]'")

    def __setitem__(self, ref, column):
        if ref not in self.headers:
            raise ValueError("column '{}' not in headers".format(ref))

        column = reduce_extra_dimen(column, 1)
        nd = num_dimen(column)

        if nd == 0:
            column = [column]
        elif nd == 2:
            column = transpose(column)[0]

        if len(column) != self.num_rows:
            raise ValueError('invalid length')

        i = self.headers[ref]
        for row, v in zip(self, column):
            row.values[i] = v

    def __iter__(self):
        return iter(islice(self._matrix, 1, None))

    def __repr__(self):
        headers = repr_iter(self.headers.keys(), wrap='[]')
        return '{} ({:,})'.format(headers, self.num_rows)






import textwrap

from collections import OrderedDict
from collections import namedtuple

from .. util.text import repr_iter


class flux_row_cls:

    def __init__(self, headers, values):
        """ basically, a mutable namedtuple

        :param headers: OrderedDict of {'header': index}
        :param values:  list of underlying data

        headers is a single dictionary passed *byref* from the
        flux_cls to all flux_row_cls instances, reducing unneccesary
        memory usage and allowing all mapping updates to be made
        instantaneously

            namedtuples may be more efficient, but too much of a headache to deal
            with their immutability

            maybe __slots__ or weakref / metaclass would be better?
            Effective Python: Item 35: Annotate Class Attributes with Metaclasses
        """
        # cannot call attributes directly in __init__ as this invokes __setattr__ lookup too early
        self.__dict__['_headers'] = headers
        self.__dict__['values']   = values

    @property
    def names(self):
        return list(self._headers.keys())

    @property
    def view_as_array(self):
        """ for development purposes; used purely to trigger a debugging feature in PyCharm

        PyCharm will recognize returned the (name, value) pairs as an ndarray
        and enable the "...view as array" option in the debugger which displays
        values in a special window
        """
        import numpy
        return numpy.transpose([self.names, self.values])

    def bind(self):
        """ bind headers / values directly to object, avoiding need for subsequent __getattr__ lookups
        ooooooooooh this is dangerous to flux_cls tho
        """
        self.__dict__.update(zip(self.names, self.values))

    def dict(self):
        return OrderedDict(zip(self.names, self.values))

    def namedtuples(self):
        nt_cls = namedtuple('flux_row_nt', self.names)
        return nt_cls(*self.values)

    def __getattr__(self, name):
        """  eg:
             v = row.header

        self.__dict__ is empty when being deserialized from a binary file
        '__setstate__' key must be called with self.__getattribute__
        """
        if not self.__dict__:
            return self.__getattribute__(name)

        return self.__getitem__(name)

    def __getitem__(self, name):
        """ eg:
            v = row['header']

        list index is translated from self._headers dictionary
        self._headers.get(name, name) is a shortcut for avoiding an isinstance(name, int) check

        i *could* handle slices here eg, row[1:], but it would just be easier
        for client to call row.values[1:] instead
        """
        try:
            i = self.__dict__['_headers'].get(name, name)
            return self.__dict__['values'][i]
        except (TypeError, IndexError):
            pass

        raise AttributeError(self.__attr_err_msg(name))

    def __setattr__(self, name, value):
        """ eg:
            row.header = v
        """
        if name in self.__dict__:               # a flux_row_cls property
            self.__dict__[name] = value
        else:                                   # a {self._headers: self.values} pair
            self.__setitem__(name, value)

    def __setitem__(self, name, value):
        """ eg:
            row['header'] = v
        """
        try:
            i = self.__dict__['_headers'].get(name, name)
            self.__dict__['values'][i] = value
            return
        except (TypeError, IndexError):
            pass

        raise AttributeError(self.__attr_err_msg(name))

    def __attr_err_msg(self, name):
        if isinstance(name, slice):
            return 'slice should be used on row.values\n(eg, row.values[2:5], not row[2:5])'

        names = repr_iter(self.names, concat=',\n', quotes=True)
        names = textwrap.indent(names, '  ')

        return "flux_row_cls\nno column named: '{}' from\n{}".format(name, names)

    def __len__(self):
        return len(self.values)

    def __iter__(self):
        return iter(self.values)

    def __repr__(self):
        return '<flux_row> ' + repr(self.values)


class lev_row_cls(flux_row_cls):

    def __init__(self, headers, values, address):
        super().__init__(headers, values)
        self.__dict__['address'] = address

    def __repr__(self):
        return self.address + ' ' + repr(self.values)




import re
from collections import Iterable

from functools import lru_cache

from datetime import date
from datetime import datetime
from datetime import timedelta

from dateutil.parser import parse as dateutil_parse


def to_datetime(v):
    """ convert variety of date representations into a datetime object
    :param v:
        an iterable, excel serial date, datetime, or string
    """

    if isinstance(v, datetime):
        d_ret = v
    elif type(v) == date:
        d_ret = __date_to_datetime(v)
    elif isinstance(v, float):
        d_ret = __parse_date_float(v)
        if d_ret is None:
            d_ret = __parse_excel_serial(v)
    elif isinstance(v, str):
        d_ret = __parse_string(v)
    elif isinstance(v, Iterable):
        d_ret = [to_datetime(d) for d in v]
    else:
        raise ValueError("invalid date: '{}', dont know how to convert '{}' instance".format(v, type(v)))

    return d_ret


def attempt_to_datetime(v):
    try:
        v = to_datetime(v)
        b = True
    except (ValueError, AttributeError):
        b = False

    return v, b


def is_valid_date(v):
    try:
        to_datetime(v)
        return True
    except (ValueError, AttributeError):
        return False


def excel_epoch():
    """ rnumber of days since 1900-01-01 """
    return datetime(1899, 12, 31)


def __parse_date_float(f):
    s = str(int(f))
    d_ret = __datetime_ymd(s)

    return d_ret


def __datetime_ymd(s):
    try:
        y, m, d = (int(s[:4]),
                   int(s[4:6]),
                   int(s[6:8]))
        return datetime(y, m, d)

    except ValueError:
        return None


def __parse_excel_serial(f):
    num_days = int(f) - 1
    return excel_epoch() + timedelta(days=num_days)


@lru_cache(maxsize=2**13)
def __parse_string(s):
    """
    time_fragment_re
        for cases like: '2017/01/01 00:00:00 000'
        these extra digits after the '00:00:00' value will break most parsers
        and should be removed
        eg:
            ' 000' = time_fragment_re.search('2017/01/01 00:00:00 000')
    """
    time_fragment_re = re.compile(r'''
        [\s][\d]{3,}$
    ''', re.X | re.I)

    s = s.strip()
    s = time_fragment_re.sub('', s)

    d_ret = __parse_string_strptime(s)
    if d_ret is None:
        d_ret = __parse_string_dateutil(s)

    return d_ret


def __parse_string_strptime(s):
    # has hh:mm value, strptime not set up to handle timestamps
    if ':' in s:
        return None

    common_formats = ('%m/%d/%Y',       # 01/01/2001
                      '%Y-%m-%d',       # 2000-01-01
                      '%Y-%b-%d',       # 2000-jan-01
                      '%d-%b-%Y',       # 01-jan-2000
                      '%Y%m%d')         # 20000101

    for d_fmt in common_formats:
        try:
            return datetime.strptime(s, d_fmt)
        except ValueError:
            pass

    return None


def __parse_string_dateutil(s):
    try:
        return dateutil_parse(s)
    except ValueError:
        return None


def __date_to_datetime(date_t):
    return datetime(date_t.year, date_t.month, date_t.day)





from . text import print_runtime
from . dates import to_datetime
from . dates import attempt_to_datetime
from . dates import is_valid_date


__all__ = ['filesystem',
           'print_runtime',
           'to_datetime',
           'attempt_to_datetime',
           'is_valid_date']




"""
    Working with Excel can be tedious.

    It's extremely easy to build terrible worksheets and there is no enforcement
    in data integrity -- anyone can just slap in some data where ever they feel like.
    Usually, this results in workbooks that are like a dirty laundry basket,
    where dats is tossed together and draped over the edges

    https://imgs.xkcd.com/comics/algorithms.png

    In VBA, uou'll spend A LOT of time constantly finding range boundaries in the sheets,
    determining last row or last column before you can start working on the data.
    It's also a big vulnerability to use arbitrary alpha column (A-Z)
    references which lead to invalid references in the VBA code whenever
    columns are shifted or reordered. Ideally, it would be nice to
    attach header names to the rows you are iterating over, much more like
    fields in a database table

    Excel's flexibility is its greatest weakness, but also its greatest strength.
    It's the best software in the world for fast prototyping and reviewing data,
    it just shouldn't be used to build entire processes. For middle-sized data,
    Excel can be phenominally powerful.

    Published site-packages like xlrd, xlwt, openpyxl will allow you to
    read Excel files from python, but don't control the actual application,
    limiting your ability to manage any existing infrastructure built on Excel

    The excel_levity_cls is meant to make Excel data feel as light as a feather
    instead of like busting concrete
"""


from time import sleep

import vengeance as veng
from examples import excel_project_template as share

xlPasteColumnWidths = 8

xlCalculationManual    = -4135
xlCalculationAutomatic = -4105

xlNone   = -4142
xlYellow = 65535
xlBlue   = 15773696
xlPink   = 9856255


@veng.print_runtime
def main():
    share.open_project_workbook(open_new_instance=True,
                                read_only=True,
                                update_links=True)

    instantiate_lev('sheet1')
    # instantiate_lev('sheet2')
    # instantiate_lev('empty sheet')
    # instantiate_lev('jagged rows')

    # lev_subsections()

    iterate_primitive_rows()
    iterate_flux_rows()
    iterate_excel_errors()

    # write_values()
    # write_values_from_lev()
    # append_values()
    # write_formulas()

    # modify_range_values(iter_method='slow')
    modify_range_values(iter_method='fast')

    # excel_object_model()
    # allow_focus()

    # share.close_project_workbook(save=False)


def instantiate_lev(tab_name):
    """
    excel_levity_cls range reference syntax:
        returns win32com reference to Excel range

        lev['<col><row>:<col><row>']
        anchor reference mnemonics:
            '*h': header
            '*f': first
            '*l': last
    """
    # help(veng.excel_levity_cls)

    lev = share.tab_to_lev(tab_name)
    a = repr(lev)
    a = lev.is_empty

    a = lev.header_r
    a = lev.first_c
    a = lev.last_c
    a = lev.first_r
    a = lev.last_r
    a = lev.append_r

    a = lev.has_headers
    a = lev.headers
    a = lev.header_values

    a = lev.named_ranges
    a = lev['some_named_range_1'].Value
    a = lev['some_named_range_2'].Value
    a = lev['excel_date'].Value

    # examples of excel_levity_cls reference syntax
    if 'col_b' in lev.headers:
        a = lev['col_b first_r : col_b last_r'].Address

    a = lev['first_c header_r : last_c last_r'].Address
    a = lev['*f *h:*l *l'].Address
    a = lev['A*h:E*l'].Address

    # or lev.sheet.Range('...') for literal addresses
    a = lev['A1'].Value
    a = lev['A1:C10'].Value

    lev['*f *h:*l *l'].Interior.Color = xlPink
    lev['*f *f:*l *l'].Interior.Color = xlBlue

    lev.clear_filter()
    lev.remove_filter()
    lev.reapply_filter()


def lev_subsections():
    """
    row 1 (colored grey) in the Excel worksheets can be used as subsection markup,
    although any other column reference will work

    eg:
        share.lev_subsection('subsections', '<sect_2>', '</sect_2>')
        share.lev_subsection('subsections', 'col_d', 'col_h')
        share.lev_subsection('subsections', 'F', 'H')
    """
    lev_1 = share.lev_subsection('subsections', '<sect_1>', '</sect_1>')
    lev_2 = share.lev_subsection('subsections', '<sect_2>', '</sect_2>')
    lev_3 = share.lev_subsection('subsections', '<sect_3/>', '<sect_3/>')

    a = lev_1.last_r
    b = lev_2.last_r
    c = lev_3.last_r

    lev_1['*f *h:*l *l'].Interior.Color = xlYellow
    lev_2['*f *h:*l *l'].Interior.Color = xlBlue
    lev_3['*f *h:*l *l'].Interior.Color = xlPink

    lev_1.clear('*f *f:*l *l', clear_colors=True)
    lev_2.clear('*f *f:*l *l', clear_colors=True)
    lev_3.clear('*f *f:*l *l', clear_colors=True)

    lev = share.tab_to_lev('subsections')
    lev['*f *h:*l *l'].Interior.Color = xlNone
    lev.reapply_filter()


def iterate_primitive_rows():
    """ iterate rows as a list of primitive values

    .rows(r_1='*h', r_2='*l'):
        * r_1, r_2 are the start and end rows of iteration
          the default values are the specialized anchor references
          starting at header row, ending at last row
        * r_1, r_2 can also be integers corresponding to the rows in the Excel range

    m = list(lev.rows())
        * as full matrix, includes header row

    m = list(lev.rows('*f'))
        * as full matrix, excludes header row

    m = lev['*f *f:*l *l'].Value2
        * returns values directly from Excel range
        * however, there may be reasons you don't want to do this,
          potentially because the matrix returns as read-only tuples,
          and because Excel's error values are not recognized (see iterate_excel_errors())
    """
    # lev = share.tab_to_lev('Sheet1')
    lev = share.tab_to_lev('errors')

    # see docstring for caveats to this
    # m = lev['*f *f:*l *l'].Value2

    for row in lev.rows():
        a = row[0]

    m = list(lev.rows('*f', 10))

    # build new matrix from filtered rows
    m = [lev.header_values]
    for r, row in enumerate(lev.rows('*f')):
        if r % 2 == 0:
            m.append(row)


def iterate_flux_rows():
    """ iterate rows as flux_row_cls objects

    .flux_rows(r_1='*h', r_2='*l'):
        * r_1, r_2 are the start and end rows of iteration
          the default values are the specialized anchor references
          starting at header row, ending at last row
        * r_1, r_2 can also be integers corresponding to the rows in the Excel range

    for row in lev:
        * preferred iteration syntax
        * begins at first row, not header row

    m = list(lev)
        * as full matrix, excludes header row

    m = list(lev.flux_rows())
        * as full matrix, includes header row
    """
    lev = share.tab_to_lev('Sheet1')

    for row in lev:
        a = row.address
        a = row.names
        a = row.values

        a = row.view_as_array       # meant as a debugging tool in PyCharm

        if 'col_a' in lev.headers:
            a = row.col_a
            a = row['col_a']
            a = row[0]

    m = list(lev.flux_rows(5, 10))

    # extract primitive values
    m = [row.values for row in lev]

    # build new matrix from filtered rows
    m = [lev.header_values]
    for r, row in enumerate(lev):
        if r % 2 == 0:
            m.append(row.values)


def iterate_excel_errors():
    lev = share.tab_to_lev('errors')

    # these will return excel's integer error code, be careful
    for row in lev['B3:D6'].Value:
        a = row[-1]

    # reading values with .rows() or .flux_rows() escapes these errors
    for row in lev.rows(3, 3):
        a = row[-1]

    for row in lev.flux_rows(3, 3):
        a = row.col_c


def write_values():
    lev = share.tab_to_lev('Sheet2')
    lev.clear('*f *f:*l *l')

    # write single value
    lev['*f *f'] = 'hello'

    # write single row
    lev['*f *f'] = ['hello', 'hello', 'hello']

    # write single column
    lev['*f *f'] = [['hello'], ['hello'], ['hello'], ['hello'], ['hello'], ['hello']]

    # write rows and columns
    m = [['col_a', 'col_b', 'col_c']]
    m.extend([['blah', 'blah', 'blah']] * 10)

    lev.clear('*f *f:*l *l')
    lev['*f *h'] = m

    # helper function shortcut
    share.write_to_tab('Sheet2', m)

    # Excel dates
    a = lev['excel_date'].Value
    a = lev['excel_date'].Value2

    # convert Excel eopoch float value into datetime
    b = veng.to_datetime(a)

    # Excel only accepts datetime.datetime, not datetime.date
    try:
        lev['excel_date'].Value = b.date()
    except TypeError as e:
        print(e)

    # the excel_levity_cls.__setitem__ protects against this issue
    lev['excel_date'] = b.date()


def write_values_from_lev():
    lev_1 = share.tab_to_lev('Sheet1')
    lev_2 = share.tab_to_lev('Sheet2')

    lev_2.clear('*f *f:*l *l')
    lev_2['*f 5'] = lev_1.rows(5, 10)

    lev_2.clear('*f *f:*l *l')
    lev_2['*f *h'] = lev_1

    # helper function
    share.write_to_tab(lev_2, lev_1)


def append_values():
    lev = share.tab_to_lev('Sheet2')

    lev.clear('*f *f:*l *l')
    a = lev.append_r

    # if sheet is empty: append_r = first_r
    m = [['a', 'a', 'a']] * 10
    lev['*f *a'] = m

    # if data is already present: append_r = last_r + 1
    a = lev.append_r

    m = [['a', 'a', 'a']] * 10
    lev['*f *a'] = m

    # or use helper function
    share.write_to_tab(lev, ['d', 'd', 'd'], r_1='*a')


def write_formulas():
    lev = share.tab_to_lev('Sheet2')
    lev.clear('*f *f:*l *l')

    lev.application.Calculation = xlCalculationManual

    # lev['col_a *f'] = '=(1 + 0)'
    # lev['col_b *f'] = '=(1 + 1)'
    # lev['col_c *f'] = '=(1 + 2)'

    # lev['col_a 4'] = '=({}{} + 10)'.format(lev.headers['col_a'], lev.first_r)
    # lev['col_b 4'] = '=({}{} + 20)'.format(lev.headers['col_b'], lev.first_r)
    # lev['col_c 4'] = '=({}{} + 30)'.format(lev.headers['col_c'], lev.first_r)

    lev['*f *h'] = ['col_a', 'col_b', 'col_c', 'col_d', 'col_e', 'col_f', 'col_h', 'col_i', 'col_j']

    lev['col_a *f'] = '=(1 + 0)'
    lev['col_a *f:*l *f'].FillRight()

    lev['col_a 4'] = '=({}{} + 10)'.format(lev.headers['col_a'], lev.first_r)
    lev['col_a 4:*l 4'].FillRight()

    lev.last_r = 20
    lev['col_a 4:*l *l'].FillDown()
    lev.calculate()


@veng.print_runtime
def modify_range_values(iter_method='slow'):
    """
    although calls like "ws.Range('...').Value" work fine in VBA,
    this is painfully slow for win32com remote procedure calls

    value extraction into flux_cls can also be used for more complex transformations
    """
    lev = share.tab_to_lev('Sheet1')
    # flux = share.tab_to_flux('Sheet1')

    if iter_method == 'slow':
        ws = lev.sheet

        for r in range(2, lev.last_r + 1):
            if r % 100 == 0:
                print('at row: {:,}'.format(r))

            ws.Range('C' + str(r)).Value = 'blah'
            if ws.Range('B' + str(r)).Value == 'find':
                ws.Range('B' + str(r)).Value = 'replace'

    else:
        m = list(lev)
        for row in m:
            if row.col_b == 'find':
                row.col_b = 'replace'

        lev['*f *f'] = m


def excel_object_model():
    from vengeance.excel_com.worksheet import activate_sheet
    from vengeance.excel_com.worksheet import clear_worksheet_filter

    ws = veng.get_worksheet(share.wb, 'object model')
    activate_sheet(ws)
    ws.Range('B2:D10').Interior.Color = xlNone

    clear_worksheet_filter(ws)

    ws.Range('B2:B10').Interior.Color = xlYellow
    ws.Range('C2:C10').Interior.Color = xlBlue
    ws.Range('D2:D10').Interior.Color = xlPink

    lev_1 = share.tab_to_lev('Sheet1')
    lev_2 = share.tab_to_lev('object model')

    lev_2.remove_filter()
    lev_2.reapply_filter()

    # special formatting
    lev_2['I *h:K *h'].ColumnWidth = 1

    lev_1['*f *h:*l *h'].Copy()
    lev_2['I *h'].PasteSpecial(Paste=xlPasteColumnWidths,
                               Operation=xlNone,
                               SkipBlanks=False,
                               Transpose=False)

    lev_1.application.CutCopyMode = False

    # invoke macro
    path = "'{}'".format(share.wb.FullName)
    macro_name = 'msgbox_test'
    share.wb.Application.Run(path + '!' + macro_name, 'hello from python')


def allow_focus():
    """
    if excel_levity_cls.allow_focus = False, lev.activate() will have no effect

    setting this value to False allows processes to run without disrupting the user
    """
    print()
    activate_all_sheets()

    print()

    veng.excel_levity_cls.allow_focus = True
    activate_all_sheets()


def activate_all_sheets():
    print('veng.excel_levity_cls.allow_focus = {}'.format(veng.excel_levity_cls.allow_focus))

    for ws in share.wb.Sheets:
        print("activate: '{}'".format(ws.Name))
        lev = share.tab_to_lev(ws, clear_filter=True)
        lev.activate()

        sleep(0.25)


main()





from time import sleep

import ctypes
import pythoncom

# noinspection PyUnresolvedReferences
from pythoncom import com_error

from ctypes import byref
from ctypes import PyDLL
from ctypes import POINTER
from ctypes.wintypes import DWORD
from ctypes.wintypes import BOOL

from win32com.client import Dispatch as pywin_dispatch
# from win32com.client.gencache import EnsureDispatch as pywin_dispatch

from comtypes import COMError
from comtypes import GUID
from comtypes import IUnknown
from comtypes.automation import IDispatch
from comtypes.client.dynamic import Dispatch
from comtypes.client import CreateObject

from .. util.text import vengeance_message
from .. util.filesystem import assert_path_exists
from .. util.filesystem import standardize_path
from .. util.filesystem import file_extension

from . excel_constants import *

AccessibleObjectFromWindow = ctypes.oledll.oleacc.AccessibleObjectFromWindow
FindWindowEx               = ctypes.windll.user32.FindWindowExA
GetWindowText              = ctypes.windll.user32.GetWindowTextA
SetForegroundWindow        = ctypes.windll.user32.SetForegroundWindow

corrupt_hwnds = set()


def open_workbook(path,
                  excel_instance=None,
                  *,
                  read_only=False,
                  update_links=True):

    wb = is_workbook_open(path)

    if wb is None:
        if not excel_instance:
            excel_app = empty_excel_instance() or new_excel_instance()
        else:
            excel_app = excel_instance

        wb = __workbook_from_excel_app(excel_app, path, update_links, read_only)

    elif excel_instance is not None:
        if wb.Application != excel_instance:
            vengeance_message("'{}' already open in another Excel instance".format(wb.Name))
            sleep(3)

    if wb.ReadOnly is False and read_only:
        vengeance_message("'{}' is NOT opened read-only".format(wb.Name))
        sleep(3)

    if wb.ReadOnly and read_only is False:
        vengeance_message("('{}' opened as read-only)".format(wb.Name))
        sleep(3)

    return wb


# noinspection PyUnusedLocal
def close_workbook(wb, save):
    """
    all references need to be severed for excel_com pointer to be released
    variables should be set to None
    """
    if save and wb.ReadOnly:
        raise AssertionError("workbook: '{}' is open read-only, cannot save and close".format(wb.Name))

    excel_app = wb.Application
    if save:
        wb.Save()
    else:
        excel_app.DisplayAlerts = False

    wb.Close()
    wb = None

    if save is False:
        excel_app.DisplayAlerts = True

    if excel_app.Workbooks.Count == 0:
        excel_app.Quit()
        excel_app = None


def is_workbook_open(path):
    """ scan all open workbooks across all Excel sessions, match is determined by identical file path """
    path = standardize_path(path)
    assert_path_exists(path)

    window_h = FindWindowEx(0, 0, xl_class_name, None)
    while window_h != 0:
        for wb in __workbooks_from_hwnd(window_h):
            path_search = standardize_path(wb.FullName)

            if path == path_search:
                return wb

        window_h = FindWindowEx(0, window_h, xl_class_name, None)

    return None


def new_excel_instance():
    excel_app = CreateObject('Excel.Application', dynamic=True)
    excel_app = __comtype_to_pywin_obj(excel_app, IDispatch)
    excel_app = pywin_dispatch(excel_app)

    # excel_app = pywin_dispatch('Excel.Application')

    excel_app.WindowState = xl_maximized
    excel_app.Visible = True

    app_to_foreground(excel_app)
    reload_all_add_ins(excel_app)

    return excel_app


def empty_excel_instance():
    window_h = FindWindowEx(0, 0, xl_class_name, None)

    while window_h != 0:
        excel_app = __excel_app_from_hwnd(window_h)
        if __is_excel_app_empty(excel_app):
            vengeance_message('utilizing empty Excel instance ...')

            return excel_app

        window_h = FindWindowEx(0, window_h, xl_class_name, None)

    return None


def any_excel_instance():
    return pywin_dispatch('Excel.Application')


# def any_excel_instance_old():
#     window_h = FindWindowEx(0, 0, xl_class_name, None)
#
#     while window_h != 0:
#         excel_app = __excel_app_from_hwnd(window_h)
#         if excel_app is not None:
#             return excel_app
#
#         window_h = FindWindowEx(0, window_h, xl_class_name, None)
#
#     return None


def __is_excel_app_empty(excel_app):
    if excel_app is None:
        return False

    workbooks = list(excel_app.Workbooks)
    if not workbooks:
        excel_app.Visible = True
        return True

    if len(workbooks) == 1:
        wb = workbooks[0]
        if wb.Saved and wb.Name == 'Book1':
            ws = wb.Sheets[1]
            if ws.Name == 'Sheet1' and ws.UsedRange.Address == '$A$1':
                return True

    return False


def __workbook_from_excel_app(excel_app, path, update_links, read_only):
    if read_only:
        vengeance_message('opening workbook as read-only ...')

    assert_path_exists(path)

    excel_app.DisplayAlerts = False
    wb = excel_app.Workbooks.Open(path, update_links, read_only)
    excel_app.DisplayAlerts = True

    return wb


def __workbooks_from_hwnd(window_h):
    excel_app = __excel_app_from_hwnd(window_h)
    if excel_app is not None:
        return list(excel_app.Workbooks)
    else:
        return []


def __excel_app_from_hwnd(window_h):
    """
    comtypes library is used to search windows handles for Excel application,
    then converts that pointer to a pywin object thru __comtype_to_pywin_obj()

    sometimes, non-Excel applications are running under the same window_h
    as an Excel process, like "print driver host for applications"

    these will fail to return a valid excel7_wnd for FindWindowEx,
    but killing these processes will also bring down the Excel application, which is
    not neccessarily corrupt
    """
    global corrupt_hwnds

    if window_h in corrupt_hwnds:
        return None

    desk_wnd   = FindWindowEx(window_h, None, xl_desk_class, None)
    excel7_wnd = FindWindowEx(desk_wnd, None, xl_excel7_class, None)

    if excel7_wnd == 0:
        corrupt_hwnds.add(window_h)
        if __is_excel_process(window_h):
            __kill_task(window_h)

        return None

    cls_id  = GUID.from_progid(xl_clsid)
    obj_ptr = ctypes.POINTER(IDispatch)()

    AccessibleObjectFromWindow(excel7_wnd,
                               native_om,
                               byref(cls_id),
                               byref(obj_ptr))
    window = Dispatch(obj_ptr)

    try:
        com_obj = window.application
        excel_app = __comtype_to_pywin_obj(com_obj, IDispatch)
        excel_app = pywin_dispatch(excel_app)

        excel_app.Visible = True

        return excel_app

    except (COMError, com_error, NameError) as e:
        raise ChildProcessError('remote procedure call to Excel application rejected\n'
                                '(check if cursor is still active within a cell somewhere, '
                                'Excel will reject automation calls while waiting on '
                                'user input)') from e


def __comtype_to_pywin_obj(ptr, interface):
    """Convert a comtypes pointer 'ptr' into a pythoncom PyI<interface> object.

    'interface' specifies the interface we want; it must be a comtypes
    interface class.  The interface must be implemented by the object;
    and the interface must be known to pythoncom.
    """
    com_obj = PyDLL(pythoncom.__file__).PyCom_PyObjectFromIUnknown

    com_obj.restype  = ctypes.py_object
    com_obj.argtypes = (ctypes.POINTER(IUnknown), ctypes.c_void_p, BOOL)

    # noinspection PyProtectedMember
    return com_obj(ptr._comobj, byref(interface._iid_), True)


def __is_excel_process(window_h):
    SysAllocStringLen           = ctypes.windll.oleaut32.SysAllocStringLen
    SysAllocStringLen.argtypes  = (ctypes.c_wchar_p, ctypes.c_uint)
    SysAllocStringLen.restype   = ctypes.POINTER(ctypes.c_char)

    chr_buffer = SysAllocStringLen(' ' * 255, 255)
    GetWindowText(window_h, chr_buffer, 255)

    name = ctypes.cast(chr_buffer, ctypes.c_char_p).value
    name = name.decode('ascii').lower()

    return name == 'excel'


def __kill_task(window_h):
    GetWindowThreadProcessId = ctypes.windll.user32.GetWindowThreadProcessId
    OpenProcess              = ctypes.windll.kernel32.OpenProcess
    TerminateProcess         = ctypes.windll.kernel32.TerminateProcess
    CloseHandle              = ctypes.windll.kernel32.CloseHandle

    vengeance_message('attempting to kill corrupt Excel application: {}'.format(window_h))

    lp_ptr = POINTER(DWORD)()
    GetWindowThreadProcessId(window_h, byref(lp_ptr))

    handle = OpenProcess(process_terminate, False, lp_ptr)
    TerminateProcess(handle, -1)
    CloseHandle(handle)


def app_to_foreground(excel_app):
    excel_app.Visible = True
    SetForegroundWindow(excel_app.Hwnd)


def add_in_exists(excel_app, name):
    try:
        excel_app.AddIns(name)
    except COMError:
        return False

    return True


def reload_all_add_ins(excel_app):
    vengeance_message('reloading Excel add-ins...')

    for add_in in excel_app.AddIns:
        if add_in.Installed:
            name = add_in.Name
            try:
                add_in.Installed = False
                add_in.Installed = True
                vengeance_message('{}'.format(name))
            except COMError:
                vengeance_message('failed to load add-in: {}' + name)
    print()


def reload_add_in(excel_app, name):
    if add_in_exists(excel_app, name):
        excel_app.addins(name).Installed = False
        excel_app.addins(name).Installed = True


def is_workbook_an_addin(f_name):
    return 'xla' in file_extension(f_name)





import copy

from collections import OrderedDict
from collections import defaultdict
from collections import Iterable
from types import GeneratorType


def reduce_extra_dimen(v, minimum=0):
    """ removes items from iterable containers if they only have a single item """
    nd = num_dimen(v)

    i = 0 + minimum
    while i < nd:
        if is_subscriptable(v) and len(v) == 1:
            v = v[0]

        i += 1

    return v


def force_two_dimen(v):
    v  = generator_to_list(v)
    nd = num_dimen(v)

    if nd > 2:
        raise ValueError('can not reduce from {} dimensions'.format(nd))

    for _ in range(2 - num_dimen(v)):
        v = [generator_to_list(v)]

    return v


# def force_two_dimen(v):
#     """
#     should one dimensional list be transposed?
#     or simply wrapped in outer list?
#     """
#     v  = generator_to_list(v)
#     nd = num_dimen(v)
#
#     if nd > 2:
#         raise ValueError('can not reduce to 2 dimensions from {} dimensions'.format(nd))
#
#     if nd == 0:
#         v = [[v]]
#     elif nd == 1:
#         v = transpose(v)
#
#     return v


def num_dimen(v):
    """ determine iteration-depth (number of dimensions)
    eg:
        0 = num_dimen('a')
        1 = num_dimen(['a'])
        1 = num_dimen([])
        2 = num_dimen(['a'], ['b']])
        2 = num_dimen([[]])
    """
    if not is_iterable(v):
        return 0

    if not is_subscriptable(v):
        return 1

    if len(v) == 0:
        return 1

    return 1 + num_dimen(v[0])


def is_iterable(v):
    """ if value is not a string, determine if it can be iterated
    eg:
        False = is_iterable('mike')
        True  = is_iterable(['m', 'i' 'k' 'e'])
    """
    is_iter = isinstance(v, Iterable)
    is_str  = isinstance(v, str)

    return is_iter and not is_str


def is_subscriptable(v):
    return isinstance(v, (list, tuple))


def generator_to_list(v):
    if is_vengeance_class(v):
        v = v.rows()

    if isinstance(v, GeneratorType):
        v = list(v)

    return v


def is_vengeance_class(o):
    bases = {b.__name__ for b in o.__class__.mro()}
    if 'flux_cls' in bases or 'excel_levity_cls' in bases:
        return True

    return False


def transpose(m):
    m = generator_to_list(m)

    if num_dimen(m) == 1:
        num_r = 1
        num_c = len(m)
        t = [[m[r] for _ in range(num_r)]
                   for r in range(num_c)]
    else:
        num_r = len(m)
        num_c = len(m[0])
        t = [[m[c][r] for c in range(num_r)]
                      for r in range(num_c)]

    return t


def append_matrices(direction, *matrices, has_header=True):
    """
    eg direction:
        'vertical',   'rows'
        'horizontal', 'columns'
    """
    # append vertically
    if direction.startswith('v') or direction.startswith('row'):
        return append_rows(*matrices, has_header=has_header)

    # append horizontally
    if direction.startswith('h') or direction.startswith('col'):
        return append_columns(*matrices)

    raise ValueError("invalid direction: '{}'".format(direction))


def append_rows(*matrices, has_header=True):

    def append(m_1, m_2):
        m_1 = force_two_dimen(m_1)
        m_2 = force_two_dimen(m_2)

        if len(m_1[0]) != len(m_2[0]):
            raise IndexError('vertical append requires matrices to have equal number of columns')

        # if either matrix is empty, return the other
        if is_empty(m_1):
            return m_2

        if is_empty(m_2):
            return m_1

        m_a = copy.copy(m_1)
        if has_header:
            m_a.extend(m_2[1:])
        else:
            m_a.extend(m_2)

        return m_a

    m_f, matrices = matrices[0], matrices[1:]
    for m in matrices:
        m_f = append(m_f, m)

    return m_f


def append_columns(*matrices):

    def append(m_1, m_2):
        m_1 = force_two_dimen(m_1)
        m_2 = force_two_dimen(m_2)

        if len(m_1) != len(m_2):
            raise IndexError('horizontal append requires matrices to have equal number of rows')

        # if either matrix is empty, return the other
        if is_empty(m_1):
            return m_2

        if is_empty(m_2):
            return m_1

        m_a = []
        for row_1, row_2 in zip(m_1, m_2):
            m_a.append(row_1 + row_2)

        return m_a

    m_f, matrices = matrices[0], matrices[1:]
    for m in matrices:
        m_f = append(m_f, m)

    return m_f


def is_empty(v):
    """ determine if iterable is composed entirely of empty iterables, eg, [] or [[]]
    eg:
        True  = is_empty( [] )
        True  = is_empty( [[]] )
        False = is_empty( [[None]] )
        False = is_empty( [[], [], []] )
    """
    m = force_two_dimen(v)
    num_rows = len(m)
    num_cols = len(m[0])

    return (num_rows == 1) and (num_cols == 0)


def make_iterable(v):
    """
    standardize values from iteration of certain datatypes
    so that the same iteration syntax can be used for multiple datatypes

    eg:
        for v in standard_iter(['abc']):
            b = 'abc'

        for v in standard_iter('abc'):
            v = 'abc'

        for v in standard_iter({'a': 0, 'b': 1, 'c': 2}):
            v = (a, 0)
    """
    if isinstance(v, (list, tuple, GeneratorType)):
        return v

    if isinstance(v, dict):
        return v.items()

    if callable(v):
        v = v()

    return [v]


def index_sequence(seq, start=0):
    items = []
    non_unique = defaultdict(int)

    for i, v in enumerate(seq, start):
        if v == '':
            v = 'None'

        if v in non_unique:
            v = '{}({})'.format(v, non_unique[v] + 1)

        items.append((v, i))
        non_unique[v] += 1

    return OrderedDict(items)


def ordered_unique(sequence):
    return list(OrderedDict((k, None) for k in sequence).keys())


def invert_mapping(d, inversion=None):
    """ return new mapping of {value: key} from {key: value}

    eg, from:
        {'a': [-1, -1, -1]
         'b': 2
         'c': 3
         'd': 3}
    to:
        {-1: ['a', 'a', 'a']
         2: b
         3: ['c', 'd']}
    """
    if not isinstance(d, dict):
        raise TypeError('value must be a dictionary')

    inverted = OrderedDefaultDict(list)
    for k, v in d.items():
        _k_, _v_ = v, k

        if inversion:
            _k_ = inversion(_k_)

        for _k_ in make_iterable(v):
            inverted[_k_].append(_v_)

    inverted = OrderedDict(inverted)
    for k, v in inverted.items():
        if len(v) == 1:
            inverted[k] = v[0]

    return inverted


class OrderedDefaultDict(OrderedDict):

    def __init__(self, default_factory=None, *a, **kw):
        if (default_factory is not None) and (not callable(default_factory)):
            raise TypeError('first argument must be callable')

        self.default_factory = default_factory

        if a and self.default_factory is list:
            self.__append_values(a[0])
        else:
            OrderedDict.__init__(self, *a, **kw)

    def __append_values(self, sequence):
        if isinstance(sequence, dict):
            sequence = OrderedDict(sequence.items()).items()
        elif not isinstance(sequence, (list, tuple)):
            raise ValueError

        for item in sequence:
            if len(item) == 2:
                self[item[0]].append(item[1])
            elif len(item) < 2:
                raise ValueError('need more than 1 value to unpack')
            elif len(item) > 2:
                raise ValueError('too many values to unpack (expected 2)')

    def __getitem__(self, key):
        try:
            return OrderedDict.__getitem__(self, key)
        except KeyError:
            return self.__missing__(key)

    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError(key)

        self[key] = value = self.default_factory()

        return value

    def defaultdict(self):
        return defaultdict(self.default_factory, self.items())

    def ordereddict(self):
        return OrderedDict(self)

    def count_values(self):
        d = OrderedDict()
        for k, v in self.items():
            d[k] = len(v)

        return d

    def reduce_extra_dimen(self, minimum=0):
        """  """
        for k, v in self.items():
            self[k] = reduce_extra_dimen(v, minimum)

            # if isinstance(v, list) and len(v) == 1:
            #     self[k] = v[0]





import json

from timeit import default_timer
from datetime import date
from collections import Iterable

nbsp = chr(160)
nonstandard_chars = "\"\'\n\t" + nbsp


def print_runtime(f):

    def runtime_wrapper(*args, **kwargs):
        f_name = f.__module__.split('.')[-1]
        f_name = '{}.{}'.format(f_name, f.__name__)

        tic = default_timer() * 1000

        if kwargs:
            result = f(*args, **kwargs)
        elif args:
            result = f(*args)
        else:
            result = f()

        toc = default_timer() * 1000
        elapsed = -(tic - toc)

        print_unicode('   τ: @{}: {}'.format(f_name, format_ms(elapsed)))

        return result

    return runtime_wrapper


def format_ms(ms):
    m, s = divmod(ms / 1000, 60)
    h, m = divmod(m, 60)

    if h >= 1:
        f_ms = '{} hr {} min'.format(h, m)
    elif m >= 1:
        f_ms = '{} min {} sec'.format(int(m), int(s))
    elif s > 1:
        f_ms = '{:.2f} sec'.format(s)
    elif ms > 10:
        f_ms = '{} ms'.format(int(ms))
    elif ms > 1:
        f_ms = '{:.1f} ms'.format(ms)
    else:
        f_ms = '{:.2f} us'.format(int(ms * 1000))

    return f_ms


def repr_iter(seq, concat=', ',
              quotes=False,
              nl=False,
              wrap=None):
    """ provide more flexibility beyond built-in repr() for iterables """

    def repr_nested(v):
        if isinstance(v, Iterable) and not isinstance(v, str):
            c = ', '
            w = None

            if is_dict_seq:
                c = ': '
            elif isinstance(v, dict):
                w = '{}'
            elif isinstance(v, tuple):
                w = '()'
            elif isinstance(v, list):
                w = '[]'

            return repr_iter(v, c, wrap=w)

        if isinstance(v, str) and quotes:
            return "'{}'".format(v)

        return str(v)

    is_dict_seq = isinstance(seq, dict)
    if is_dict_seq:
        seq = seq.items()

    if nl:
        concat = concat.strip(' ') + '\n'

    s = concat.join(repr_nested(o) for o in seq)

    if wrap:
        s = wrap[0] + s + wrap[1]

    return s


def sanitize(v):
    s = value_to_string(v)
    s = remove_multiple(s, nonstandard_chars)
    s = s.strip()

    return s


def value_to_string(v):
    if isinstance(v, str):
        return v

    if v is None:
        v = ''
    elif isinstance(v, date):
        v = v.strftime('%Y-%m-%d')
    else:
        v = str(v)

    return v


def remove_multiple(s, substrs, first_only=False):
    for sub in substrs:
        if first_only:
            s = s.replace(sub, '', 1)
        else:
            s = s.replace(sub, '')

    return s


def between(s, substr_1, substr_2=None):
    """ text between inner and outer substrings (exclusive)
    eg:
        'ell' = between('hello', 'h', 'o')
    """
    if substr_2 is None:
        substr_2 = substr_1

    i_1 = s.index(substr_1) + len(substr_1)
    i_2 = s.index(substr_2, i_1)

    return s[i_1:i_2]


def p_json_dumps(o, indent=4):
    """
    json can not convert certain python objects to string representations
        dates
        sets
    """
    def unhandled_conversion(_o):
        if isinstance(_o, date):
            return _o.isoformat()

        if isinstance(_o, set):
            return list(_o)

        raise TypeError('cannot convert type to json ' + repr(_o))

    return json.dumps(o, indent=indent,  default=unhandled_conversion)


def to_ascii(s):
    greek_chrs = {'α': 'a',
                  'β': 'b',
                  'ε': 'e',
                  'τ': 't',
                  'ν': 'v'}

    s = str(s)
    for old, new in greek_chrs.items():
        s = s.replace(old, new)

    s = (s.encode('ascii', errors='backslashreplace')
          .decode('ascii'))

    return s


def print_unicode(s):
    try:
        print(s)
    except UnicodeEncodeError:
        print(to_ascii(s))


def vengeance_message(s):
    print_unicode('   ν: {}'.format(s))






from . excel_levity_cls import excel_levity_cls
from . flux_cls import flux_cls
from . flux_row_cls import flux_row_cls

__all__ = ['excel_levity_cls',
           'flux_cls',
           'flux_row_cls']




# noinspection PyUnresolvedReferences
from pythoncom import com_error

from datetime import date
from datetime import datetime

from .. util.text import vengeance_message
from .. util.iter import force_two_dimen
from .. util.iter import is_iterable
from .. util.iter import is_vengeance_class

from .. excel_com.excel_address import col_letter
from .. excel_com.excel_address import col_number
from .. excel_com.excel_address import max_str_len
from .. excel_com.excel_address import max_cols

from .workbook import app_to_foreground
from .excel_constants import *


def get_worksheet(wb, tab_name,
                  *,
                  clear_filter=False,
                  activate=False):

    if tab_name.__class__.__name__ in {'CDispatch', '_Worksheet'}:
        return tab_name

    if wb is None:
        raise AssertionError("Cannot retrieve tab: '{}' Excel workbook has not been initialized!".format(tab_name))

    ws = None
    if isinstance(tab_name, str):
        for s in wb.Sheets:
            if s.Name.lower() == tab_name.lower():
                ws = s
                break

    elif isinstance(tab_name, int):
        ws = wb.sheets[tab_name]

    if ws is None:
        raise NameError("tab: '{}' not found in workbook: '{}'".format(tab_name, wb.name))

    ws.Visible = True

    if clear_filter:
        clear_worksheet_filter(ws)

    if activate:
        app_to_foreground(ws.Application)
        ws.Activate()

    return ws


def clear_worksheet_filter(ws):
    if ws.AutoFilter is not None:
        if ws.AutoFilter.FilterMode is True:
            # noinspection PyProtectedMember
            ws._AutoFilter.ShowAllData()


def is_range_empty(excel_range):
    num_cells = int(excel_range.Application.WorksheetFunction.CountA(excel_range))
    return num_cells == 0


def activate_sheet(ws):
    app_to_foreground(ws.Application)
    ws.Visible = True
    ws.Activate()


def write_to_excel_range(v, excel_range):
    m = force_two_dimen(v)
    m = __excel_friendly_matrix(m)

    a = excel_range.Address
    if ':' not in a:
        a = '{}:{}'.format(a, excel_range.Resize(len(m), len(m[0])).Address)
    else:
        a = excel_range.Resize(len(m), len(m[0])).Address

    excel_range.Parent.Range(a).Value = m


def gen_range_rows(excel_range):
    for row in __convert_excel_errors(excel_range):
        yield row


def __convert_excel_errors(excel_range):
    m = force_two_dimen(excel_range.Value2)
    m = [list(row) for row in m]

    try:
        range_errors = excel_range.SpecialCells(xl_cell_type_formulas, xl_errors)
        cell = first_cell(excel_range)
        r_0  = cell.Row
        c_0  = cell.Column

        for cell_error in range_errors:
            r = cell_error.Row - r_0
            c = cell_error.Column - c_0
            m[r][c] = excel_errors.get(m[r][c], 'unknown error')

    except com_error:
        pass

    return m


def __excel_friendly_matrix(m):
    """
    convert objects and datetime.dates in matrix so they can be written to an excel range
    truncate any strings that are too large
    """

    def is_primitive():
        return not hasattr(v, '__dict__')

    def is_string_too_long():
        too_long = isinstance(v, str) and len(v) > max_str_len
        if too_long:
            vengeance_message("making excel-friendly value from very long string: '{} ... {}'"
                              .format(v[:10], v[-10:]))

        return too_long

    if len(m[0]) > max_cols:
        raise ValueError("number of matrix columns ({:,}) exceeds Excel's column limit\n"
                         "(did you mean to transpose this matrix?)"
                         .format(len(m[0])))

    em = []
    for row in m:
        row_e = []

        for v in row:
            if type(v) == date:
                v = datetime(v.year, v.month, v.day)
            elif is_string_too_long():
                v = v[:32767]
            elif not is_primitive():
                v = repr(v)
            elif is_iterable(v):
                v = str(v)

            row_e.append(v)

        em.append(row_e)

    return em


def range_find(excel_range,
               what,
               after=None,
               look_in=xl_values,
               look_at=xl_whole,
               search_order=xl_by_rows,
               search_direction=xl_previous,
               match_case=False):

    if after is None:
        if search_direction == xl_next:
            after = last_cell(excel_range)
        elif search_direction == xl_previous:
            after = first_cell(excel_range)

    args = (what, after, look_in, look_at, search_order, search_direction, match_case)
    search = excel_range.Find(*args)

    return search


def first_cell(excel_range):
    return excel_range.Cells(1)


def last_cell(excel_range):
    """ excel_range.Cells(excel_range.Cells.Count) causes an overflow error
    this is not always reliable when excel_range.Cells() requires both row and column arguments
    """
    a = excel_range.Address.split(':')[-1]
    return excel_range.Parent.Range(a)


def first_row(excel_range, default_r=1):
    search = range_find(excel_range,
                        what='*',
                        search_order=xl_by_rows,
                        search_direction=xl_next)

    if search:
        r = max(search.Row, default_r)
    else:
        r = default_r

    return r


def last_row(excel_range, default_r=1):
    search = range_find(excel_range,
                        what='*',
                        search_order=xl_by_rows,
                        search_direction=xl_previous)

    if search:
        r = max(search.Row, default_r)
    else:
        r = default_r

    return r


def first_col(excel_range, default_c='A'):
    search = range_find(excel_range,
                        what='*',
                        search_order=xl_by_columns,
                        search_direction=xl_next)

    if search:
        c = max(search.Column, col_number(default_c))
        c_lttr = col_letter(c)
    else:
        c_lttr = default_c

    return c_lttr


def last_col(excel_range, default_c='A'):
    search = range_find(excel_range,
                        what='*',
                        search_order=xl_by_columns,
                        search_direction=xl_previous)

    if search:
        c = max(search.Column, col_number(default_c))
        c_lttr = col_letter(c)
    else:
        c_lttr = default_c

    return c_lttr


def first_non_blank_col(excel_range, default_c='A'):
    search = range_find(excel_range,
                        what='',
                        search_order=xl_by_columns,
                        search_direction=xl_next)

    if search:
        c = max(search.Column - 1, col_number(default_c))
        c_lttr = col_letter(c)
    else:
        c_lttr = default_c

    return c_lttr


def parse_range(excel_range):

    c_1 = first_cell(excel_range).Column
    c_1 = col_letter(c_1)

    c_2 = last_cell(excel_range).Column
    c_2 = col_letter(c_2)

    r_1 = first_cell(excel_range).Row
    r_2 = last_cell(excel_range).Row

    return c_1, r_1, c_2, r_2





"""
    Managing tabular data shouldn't be complicated.

    If values are stored in a matrix, it should't be any harder to iterate or modify
    than a normal list. One enhancement, however, would be to have row values accessible
    by column names instead of by integer indeces
        eg,
            for row in m:
                row.header

            for row in m:
                row[17]

    Two naive solutions for this are:
        1) convert rows to a dictionaries
        Using duplicate dictionary instances for every row has a high memory
        footprint, and makes accessing values by index more complicated
        eg,
            [{'col_a': 1.0, 'col_b': 'b', 'col_c': 'c'},
            {'col_a': 1.0, 'col_b': 'b', 'col_c': 'c'}]

        2) convert rows to namedtuples
        Named tuples do not have per-instance dictionaries, so they are
        lightweight and require no more memory than regular tuples,
        but their values are read-only

    Another possibility would be to store the values in column-major order,
    like in a database. This has a further advantage in that all values
    in the same column are usually of the same data type, allowing them to
    be stored more efficiently
    eg,
        row-major order
            [['coi_a', 'col_b', 'col_c'],
             [1.0,     'b',    'c'],
             [1.0,     'b',    'c'],
             [1.0,     'b',    'c']]

        column-major order
             {'col_a': [1.0, 1.0, 1.0],
              'col_a': ['b', 'b', 'b'],
              'col_a': ['c', 'c', 'c']}

    This is essentially what a pandas DataFrame is. The drawback to this
    is a major conceptual overhead.
    ***********************************************************************************
    *   Intuitively, each row is some entity, each column is a property of that row   *
    ***********************************************************************************
        * the first thing everyone looks up for a DataFrame is "how to iterate rows",
          the first thing the documentation says is "I hope you never have to use this"

        * DataFrames have some great features, but also require specialized
          syntax that can get very awkward and requires a lot of memorization


    The flux_cls attempts to balance intuitive iteration with performance

    it has the following attributes:
        * row-major order
        * named attributes on rows (that are efficiently updated)
        * value mutability on rows
        * light memory footprint
"""

from string import ascii_lowercase as ascii_s
from random import choice

import vengeance
from vengeance import flux_cls
from examples import excel_project_template as share


@vengeance.print_runtime
def main():
    flux = instantiate_flux()

    # restricted_header_names()

    # write_to_file(flux)
    # read_from_file()

    # read_from_excel()
    # write_to_excel(flux)

    # modify_columns(flux)
    # modify_rows(flux)

    iterate_primitive_rows(flux)
    iterate_flux_rows(flux)

    # flux_sort_filter(flux)
    # flux_mapping()

    # flux_subclass()

    # compare_against_pandas()


def instantiate_flux():
    # help(flux_cls)

    m = [['col_a', 'col_b', 'col_c']]
    for _ in range(100):
        m.append([''.join(choice(ascii_s) for _ in range(15))
                                          for _ in range(3)])
    flux = flux_cls(m)

    a = flux.headers
    a = flux.header_values

    a = flux.num_cols
    a = flux.num_rows

    a = flux.is_jagged

    return flux


def restricted_header_names():
    """
    the flux_row_cls has certain reserved attributes (see iterate_flux_rows())

    for row in flux:
        row.names
        row.values
    """
    try:
        m = ['names', 'col_a']
        flux = flux_cls(m)
    except NameError as e:
        print(e)

    print()

    try:
        m = ['values', 'col_a']
        flux = flux_cls(m)
    except NameError as e:
        print(e)

    m = [['names_', 'values_', 'col_a'],
         ['a', 'b', 'c']]
    flux = flux_cls(m)

    row = flux[1]
    print(row)
    a = row.names       # all flux_row column names
    b = row.names_      # value in column named "names_"
    c = row.col_a       # value in column named "col_a"

    a = row.values      # all flux_row values
    b = row.values_     # value in column named "values_"
    c = row.col_a       # value in column named "col_a"


def read_from_file():
    """ thse are CLASS METHODS, called from flux_cls not flux """
    f_dir = share.project_dir

    flux = flux_cls.from_csv(f_dir + 'flux_file.csv')
    flux = flux_cls.from_json(f_dir + 'flux_file.json')
    flux = flux_cls.deserialize(f_dir + 'flux_file.flux')


def write_to_file(flux):
    """ thse are INSTANCE METHODS, called from flux not flux_cls """
    f_dir = share.project_dir

    flux.to_csv(f_dir + 'flux_file.csv')
    flux.to_json(f_dir + 'flux_file.json')
    flux.serialize(f_dir + 'flux_file.flux')


def read_from_excel():
    share.open_project_workbook(read_only=True)
    flux = share.tab_to_flux('Sheet2')


def write_to_excel(flux):
    share.open_project_workbook(read_only=True)
    share.write_to_tab('Sheet2', flux)


def modify_columns(flux):
    """
    the flux_cls does not allow for addition / insertion an entire vector of columns yet
    instead, a call to flux.append_columns(header) must be followed by flux[header] = column
    """
    flux = flux.copy(deep=True)

    flux.rename_columns({'col_a': 'renamed_a',
                         'col_b': 'renamed_b'})

    flux.insert_columns((0,          'insert_c'),
                        ('insert_c', 'insert_b'),
                        ('insert_b', 'insert_a'),
                        ('col_c',    'insert_d'))

    flux.append_columns('append_a', 'append_b', 'append_c')

    # new column values are initialized to None (see docstring)
    col = flux['insert_a']
    col = flux['append_a']

    flux.delete_columns('insert_a', 'insert_b', 'insert_c', 'insert_d')
    flux.rename_columns({'renamed_a': 'col_a',
                         'renamed_b': 'col_b'})

    # encapsulate insertion, deletion and rename within single function
    flux.matrix_by_headers('col_c',
                           '(insert_a)',
                           {'col_a': 'renamed_a'})

    flux.fill_jagged_columns()

    # column assignment / retrieval
    flux['col_a'] = [None] * flux.num_rows
    col = flux['col_b']


def modify_rows(flux):
    flux_1 = flux.copy()
    flux_2 = flux.copy()

    flux_2.append_rows([['a', 'b', 'c']] * 25)
    a = flux_1.num_rows
    b = flux_2.num_rows

    flux_2.insert_rows(5, [['blah', 'blah', 'blah']] * 10)

    # inserting rows at index 0 will overwrite existing headers
    flux_2.insert_rows(0, [['col_d', 'col_e', 'col_f']] +
                          [['d', 'e', 'f']] * 3)
    a = flux_1.header_values
    b = flux_2.header_values

    # add rows from another flux_cls
    flux_2.insert_rows(0, flux_1)
    flux_2.append_rows(flux_1[10:15])


def iterate_primitive_rows(flux):
    """ iterate rows as a list of primitive values

    .rows(r_1='*h', r_2='*l'):
        * r_1, r_2 are the start and end rows of iteration
          the default values are the specialized anchor references
          starting at header row, ending at last row
        * r_1, r_2 can also be integers

    m = list(flux.rows())
        * as full matrix, includes header row

    m = list(flux.rows(1))
        * as full matrix, excludes header row
    """
    for row in flux.rows(3):
        a = row[0]

    m = list(flux.rows(5, 10))

    # build new matrix from filtered rows
    m = [flux.header_values]
    for r, row in enumerate(flux.rows(1), 1):
        if r % 2 == 0:
            m.append(row)

    # extract single column
    b = [row[0] for row in flux.rows()]


def iterate_flux_rows(flux):
    """ iterate rows as flux_row_cls objects

    .flux_rows(r_1='*h', r_2='*l'):
        * r_1, r_2 are the start and end rows of iteration
          the default values are the specialized anchor references
          starting at header row, ending at last row
        * r_1, r_2 can also be integers

    for row in flux:
        * preferred iteration syntax
        * begins at first row, not header row

    m = flux[i_1:i_2]
        * preferred over flux.flux_rows(i_1, i_2)
        * as full matrix, slice syntax

    m = flux[::10]
        * every 10th row

    m = list(flux.flux_rows())
        * as full matrix, includes header row

    m = list(flux)
        * as full matrix, excludes header row

    offset comparisions can easily be achieved by:
        rows = iter(flux)
        row_prev = next(rows)

        for row in rows:
            if row_prev.col_a ... and row.col_a ...:

            row_prev = row

    flux.bind()
        * speeds up attribute access on flux_row_cls, but this interferes with
          header synchronization with parent flux_cls
    """

    for row in flux:
        # see restricted_header_names()
        a = row.names
        a = row.values

        a = row.view_as_array       # meant as a debugging tool in PyCharm

        a = row.col_a
        a = row['col_a']
        a = row[0]

    # slice, stride
    m = flux[5:-2]
    m = flux[::10]

    # extract single column
    col = [row.col_b for row in flux]
    col = flux['col_b']

    # extract primitive values
    m = [row.values for row in flux]

    # build new matrix from filtered rows
    m = [flux.header_values]
    for r, row in enumerate(flux):
        if r % 2 == 0:
            m.append(row.values)


def flux_sort_filter(flux):
    # filter_function
    def starts_with_a(_row):
        return (_row.col_a.startswith('a')
                or _row.col_b.startswith('a')
                or _row.col_c.startswith('a'))

    flux = flux.copy()

    # modifications as new flux_cls
    flux_2 = flux.sorted('col_a', 'col_b', 'col_c',
                         reverse=[True, True, True])
    flux_2 = flux.filtered(starts_with_a)
    flux_2 = flux.filtered_by_unique('col_a', 'col_b')

    # in-place modifications
    flux.sort('col_a', 'col_b', 'col_c',
              reverse=[True])
    flux.filter(starts_with_a)
    flux.filter_by_unique('col_a', 'col_b')


def flux_mapping():
    """
    note there is a subtle difference in names between these functions
        flux.index_row  (singular)
        flux.index_rows (plural)

        .index_row
            * overwrites non-unique values

        .index_rows
            * will append all rows as a list to its values
            * effectively, a groupby statement
    """
    m = [['name_a', 'name_b', 'val_a', 'val_b']]
    m.extend([['a', 'b', 10, 20]] * 10)
    m.extend([['c', 'd', 50, 60]] * 20)
    flux = flux_cls(m)

    a = flux.unique_values('name_a')
    a = flux.unique_values('name_a', 'name_b')
    a = flux.namedtuples()

    d_1 = flux.index_row('name_a', 'name_b')
    d_2 = flux.index_rows('name_a', 'name_b')

    # notice the difference between these dictionaries
    a = d_1.__class__.__name__
    b = d_2.__class__.__name__

    k = ('a', 'b')
    a = d_1[k]          # .index_row  (singular)
    b = d_2[k]          # .index_rows (plural)

    # .index_rows() can act as a groupby / sumif
    for k, rows in d_2.items():
        v = sum(row.val_a + row.val_b for row in rows)
        print(k, v)


def flux_subclass():
    """
    the flux_custom_cls.commands variable is intended to provide a high-level description
    of the behaviors of this class, and making its state transformations explicit and modular

    flux.execute_commands(profile_performance=True)
        output performance metrics for each command (requires line_profiler site-package)
        very useful for debugging any performance issues for custom flux methods
    """
    m = [['transaction_id', 'name', 'apples_sold', 'apples_bought'],
         ['001', 'alice', 2, 0],
         ['002', 'alice', 0, 1],
         ['003', 'bob',   2, 1],
         ['004', 'chris', 2, 1],
         ['005', None,    7, 1]]

    flux = flux_custom_cls(m)

    flux.execute_commands(flux.commands)
    # flux.execute_commands(flux.commands, profile_performance=True)


class flux_custom_cls(flux_cls):

    # "behavior manifest" variable
    commands = ['_replace_nones',
                '_count_unique_names']

    def __init__(self, matrix):
        self.num_unique_names = []
        super().__init__(matrix)

    def _replace_nones(self):
        for row in self:
            if row.name is None:
                row.name = 'unknown'

    def _count_unique_names(self):
        self.num_unique_names = len(self.unique_values('name'))


def compare_against_pandas():
    """
    what are the tradeoffs with pandas?

    DataFrame Pros:
        performance:
            * pandas DataFrames are much more optimized for speed and memory efficiency

    flux_cls Pros:
        built on pure-python lists:
            * Python's built-in list data structure is fantastic, and is much
              less of a hassle than numpy.ndarrays or pandas.Series

            * Python's built-in list has no major inefficieny (such as appending rows to DataFrames)

            * although slower than pandas, iteration of flux_cls is every bit as fast as
              iteration over any native python list

            * if you ever HAVE to iterate rows in a DataFrame, the flux_cls can be
              up to 70x to 7,000x faster depending on whether you use .iterrows
              or .itertuples

        clarity:
            * row-major iteration coincides with the intuitive organization of the data

            * allows for much clearer syntax; the optimization in the DataFrame
              can lead developers to use difficult to read transformations, like this:
                  df['c'] = df.apply(
                      lambda row: row['a']*row['b'] if np.isnan(row['c']) else row['c'],
                      axis=1
                  )
    """
    import timeit
    import pandas

    df = None
    flux = None

    def init():
        nonlocal df
        nonlocal flux

        m = [['col_a', 'col_b', 'col_c']]
        for i in range(10000):
            if i % 10 == 0:
                m.append([None, 'fill', 'missing'])
            else:
                m.append(['blah', 'blah', 'blah'])

        df = pandas.DataFrame(m[1:], columns=m[0])

        flux = flux_cls(m)
        flux.bind()

    def as_dataframe():
        nonlocal df

        def fill(_row):
            if _row[0] is None:
                return ' '.join((_row[1], _row[2]))

        for i, row in df.iterrows():
            pass

        # for row in df.itertuples():
        #     pass

        # df['col_a'] = 1
        # df['col_a'] = df['col_b'] + df['col_c']
        # df['col_a'] = df.apply(fill)

        # a = df[df['col_a'].notnull() & (df['col_b'] == 'blah')]

        # df_2 = pandas.DataFrame([[1, 2, 3]] * 5000)
        # df.append(df_2)

    def as_flux():
        nonlocal flux

        def f(_row):
            return _row.col_a is not None and _row.col_b == 'blah'

        for row in flux:
            pass

            # row.col_a = 1
            # row.col_a = row.col_b + row.col_c
            # if row.col_a is None:
            #     row.col_a = ' '.join((row.col_b, row.col_c))

        # a = flux.filtered(f)

        # flux.append_rows([[1, 2, 3]] * 5000)

    init()

    # as_dataframe()
    # as_flux()

    t_1 = timeit.timeit(as_dataframe, number=10)
    t_2 = timeit.timeit(as_flux, number=10)
    print('as_dataframe: {:.4f}'.format(t_1))
    print('as_flux:      {:.4f}'.format(t_2))

    if t_2 < t_1:
        print('flux faster by: {:,.2f}x'.format(t_1 / t_2))
    else:
        print('dataframe faster by: {:,.2f}x'.format(t_2 / t_1))


main()





import os

import setuptools
from setuptools import setup

from textwrap import dedent


__version__  = '1.0.7'
__release__ = '$release 2'

dependencies = ('comtypes',
                'pypiwin32',
                'python-dateutil',
                'pyodbc')


long_description = '''
    need to fill this out later
'''


def write_version_file():
    f_dir = os.path.realpath(__file__)
    f_dir = os.path.split(f_dir)[0]

    with open(f_dir + '\\vengeance\\version.py', 'w') as f:
        s = '''
            # generated from setup.py
            __version__ = '{}'
            __release__ = '{}'

            dependencies = {}
        '''.format(__version__, __release__, repr(dependencies))
        
        f.write(dedent(s))


if __name__ == '__main__':
    write_version_file()

    setup(name='vengeance',
          version=__version__,
          description='Data library focusing on pure python data structures and Excel interaction',
          long_description=long_description,
          url='https://github.com/michael-ross-ven/vengeance',
          author='Michael Ross',
          author_email='michael.ross.uncc@gmail.com',
          license='MIT',
          install_requires=dependencies,
          # packages=('vengeance',),
          packages=setuptools.find_packages(),
          classifiers=[
              "Programming Language :: Python :: 3",
              "License :: OSI Approved :: MIT License",
              "Operating System :: Microsoft :: Windows"
          ]

          )
